import { _decorator, assetManager, path } from 'cc';
const { ccclass, property } = _decorator;

@ccclass('WebZipLoader')
export class WebZipLoader {
    private static _instance: WebZipLoader = null;
    public static get instance() {
        if (!this._instance) {
            this._instance = new WebZipLoader();
        }
        return this._instance;
    }



    private zipCache = new Map<string, any>();
    private resCache = new Map();

    constructor() {
        this.injectXMLHttpRequest();
        // this.injectDownloadImage('.png');
        // this.injectDownloadImage('.jpg');
        // this.injectDownloadImage('.bmp');
        // this.injectDownloadImage('.jpeg');
        // this.injectDownloadImage('.gif');
        // this.injectDownloadImage('.ico');
        // this.injectDownloadImage('.tiff');
        // this.injectDownloadImage('.webp');
        // this.injectDownloadImage('.image');

        assetManager.downloader.register({
            ".png": this.downloadImage.bind(this),
            ".jpg": this.downloadImage.bind(this),
            ".bmp": this.downloadImage.bind(this),
            ".jpeg": this.downloadImage.bind(this),
            ".gif": this.downloadImage.bind(this),
            ".ico": this.downloadImage.bind(this),
            ".tiff": this.downloadImage.bind(this),
            ".webp": this.downloadImage.bind(this),
            ".image": this.downloadImage.bind(this),
            'bundle': this.downloadBundle.bind(this)
        });


    }
    /**
  * Support the file formate was dwonloaded using XMLHttpRequest, like .cconb, .astc, .json.
  * @returns 
  */
    private injectXMLHttpRequest(): boolean {


        const that = this;

        // const originalSrcDescriptor = Object.getOwnPropertyDescriptor(HTMLScriptElement.prototype, 'src');

        // Object.defineProperty(HTMLScriptElement.prototype, 'src', {
        //     get() {
        //         return originalSrcDescriptor?.get?.call(this);
        //     },
        //     set(value) {
        //         // 检查缓存
        //         if (zipCache.has(value)) {
        //             console.error('[zipCache.js]', value);
        //             const scriptContent = zipCache.get(value)
        //             const blob = new Blob([scriptContent.async('arraybuffer')], { type: 'application/javascript' });
        //             const blobUrl = URL.createObjectURL(blob);
        //             // 延迟设置 Blob URL
        //             setTimeout(() => {
        //                 originalSrcDescriptor?.set?.call(this, blobUrl);
        //             }, 50);

        //             // 清理内存
        //             this.addEventListener('load', () => URL.revokeObjectURL(blobUrl));
        //             return;
        //         }
        //         // 无缓存则正常加载
        //         originalSrcDescriptor?.set?.call(this, value);
        //     },
        //     configurable: true,
        // });


        const accessor = Object.getOwnPropertyDescriptor(XMLHttpRequest.prototype, 'response');
        if (accessor === undefined)
            return false;

        // 注入新 open 方法
        const oldOpen = (XMLHttpRequest.prototype).open;
        // @ts-ignore
        (XMLHttpRequest.prototype).open = function (method, url: any, async, user, password) {
            const extension = url.split('.').pop();
            // 檢查zip 加載時是否有這個url
            if (that.zipCache.has(path.basename(url))) {
                this.zipCacheUrl = path.basename(url);
            } else {
                this.zipCacheUrl = null;
            }

            return oldOpen.apply(this, [method, url, async, user, password]);
        }

        // 注入新 send 方法
        const oldSend = (XMLHttpRequest.prototype).send;
        (XMLHttpRequest.prototype).send = async function (data) {
            if (this.zipCacheUrl) {
                // if (!resCache.has(this.zipCacheUrl)) {
                //     const cache = zipCache.get(this.zipCacheUrl);
                //     if (cache) {
                //         if (this.responseType === "json") {
                //             const text = await cache.async("text");
                //             resCache.set(this.zipCacheUrl, text);
                //         } else {
                //             const res = await cache.async(this.responseType);
                //             resCache.set(this.zipCacheUrl, res);
                //         }
                //         // zipCache.delete(this.zipCacheUrl);
                //     }
                // }

                if (typeof this.onload === "function") {
                    const event = new ProgressEvent('load', {
                        lengthComputable: false,
                        loaded: 0,
                        total: 0
                    });
                    this.onload(event);
                }
                return;
            }
            return oldSend.apply(this, [data]);
        }

        Object.defineProperty((XMLHttpRequest.prototype), 'response', {
            get: function () {
                if (this.zipCacheUrl) {
                    const res = that.zipCache.get(this.zipCacheUrl);
                    return this.responseType === "json" ? JSON.parse(res) : res;
                }
                return accessor.get ? accessor.get.call(this) : undefined;
            },
            set: function () { },
            configurable: true
        });

        return true;
    }


    get JSZip() {
        return new window["JSZip"]()
    }

    downloadZip(url: string): Promise<void> {
        return new Promise((resolve, reject) => {
            fetch(`${url}`)
                .then(response => {
                    if (!response.ok) {
                        throw new Error(`HTTP error! Status: ${response.status}`);
                    }
                    return response.arrayBuffer();
                })
                .then(buffer => {
                    return this.JSZip.loadAsync(buffer); // 依赖JSZip返回Promise
                })
                .then(async zipFile => {
                    // // 遍历ZIP文件并写入缓存
                    // zipFile.forEach((relativePath, fileEntry) => {
                    //     if (fileEntry.dir) return;
                    //     this.zipCache.set(`${url}/${bundleName}/${relativePath}`, fileEntry);
                    // });
                    let keys = Object.keys(zipFile.files);
                    for (let i = 0; i < keys.length; i++) {
                        let key = keys[i];
                        let data = zipFile.files[key];
                        let extName = key.split('.').pop();
                        let tempImageName = key.split('/').pop();
                        let asyncType = "";
                        switch (extName) {
                            case 'png':
                            case 'jpg':
                            case 'bmp':
                            case 'jpeg':
                            case 'gif':
                            case 'ico':
                            case 'tiff':
                            case 'webp':
                            case 'image':
                                // asyncType = sys.capabilities.imageBitmap && macro.btm ? 'blob' : 'base64';
                                asyncType = 'blob';
                                await data.async(asyncType)
                                    .then((result) => {
                                        this.zipCache.set(tempImageName, result);
                                    });
                                break;
                            case 'js':
                            case 'json':
                                asyncType = 'string';
                                await data.async(asyncType).then((result) => {
                                    this.zipCache.set(tempImageName, result);
                                });
                                break;
                            case 'default':
                                break;
                        }
                    }
                    resolve(); // 所有操作完成
                })
                .catch(e => {
                    console.error(`[${this.constructor.name}].downloadZip`, e);
                    reject(e); // 明确传递错误
                });
        });
    }

    orginalLoader(url, options, onComplete) {
        // const allowImageBitmap = sys.capabilities.imageBitmap && macro.ALLOW_IMAGE_BITMAP;
        const allowImageBitmap = true;
        if (allowImageBitmap) {
            options.responseType = 'blob';
            assetManager.downloader.downloadFile(url, options, null, (error, data) => {
                if (error) {
                    onComplete && onComplete(null, allowImageBitmap ? "" : new Image());
                } else {
                    onComplete && onComplete(null, data);
                }
            });
        } else {
            assetManager.downloader.downloadDomImage(url, options, (error, data) => {
                if (error) {
                    onComplete && onComplete(null, allowImageBitmap ? "" : new Image());
                } else {
                    onComplete && onComplete(null, data);
                }
            });
        }
    }

    downloadImage(url, options, onComplite) {

        let fileName = path.basename(url);
        if (this.zipCache.has(fileName)) {
            console.log('load file from zipFile:', fileName);
            onComplite && onComplite(null, this.zipCache.get(fileName));
        } else {
            this.orginalLoader(url, options, onComplite);
        }
    }

    // 动态执行 JavaScript 代码
    executeJSFile(code) {
        const script = document.createElement('script');
        script.textContent = code;
        document.body.appendChild(script);
    }

    REGEX = /^(?:\w+:\/\/|\.+\/).+/;

    downloadBundle(nameOrUrl, options, onComplete) {

        let bundleName = path.basename(nameOrUrl);
        let url = nameOrUrl;


        if (!this.REGEX.test(url)) url = 'assets/' + bundleName;

        var version = options.version || assetManager.downloader.bundleVers[bundleName];

        var count = 0;

        // var config = `${url}/config.${version ? version + '.' : ''}json`;
        var config = `config.${version ? version + '.' : ''}json`;
        let out = null, error = null;
        if (this.zipCache.has(config)) {
            url = '';
            count++;
            config = `config.${version ? version + '.' : ''}json`;
            out = JSON.parse(this.zipCache.get(config));
            out.base = "";
            if (count === 2) {
                onComplete(null, out);
            }
        } else {
            // url = 'assets/' + bundleName;正常的bundle加载
            config = `${url}/config.${version ? version + '.' : ''}json`;
            options.xhrResponseType = 'json';
            assetManager.downloader.downloadFile(config, options, options.onFileProgress, function (err, response) {
                if (err) {
                    error = err;
                }
                out = response;
                out && (out.base = url + '/');
                count++;
                if (count === 2) {
                    onComplete(error, out);
                }
            });
        }

        var js = `index.${version ? version + '.' : ''}js`;
        if (this.zipCache.has(js)) {
            count++;
            this.executeJSFile(this.zipCache.get(js));
            if (count === 2) {
                onComplete(null, out);
            }
        } else {
            // url = 'assets/' + bundleName; 正常的bundle加载
            js = `${url}/index.${version ? version + '.' : ''}js`;
            options.xhrResponseType = 'js';
            assetManager.downloader.downloadScript(js, options, function (err) {
                if (err) {
                    error = err;
                }
                count++;
                if (count === 2) {
                    onComplete(error, out);
                }
            });
        }

    }

}


